
/*
 * Questions or comments about this code : rao at labri dot fr
 */

#include "utils.hpp"

#ifndef NODOUBLECHECK
#define DOUBLECHECK
#endif

int nexteven(int a)
{
  if(a<2) return 2;
  return a+(a&1);
}

int next1mod4(int a)
{
  if(a>(1+(a/4*4))) return 5+(a/4*4);
  return 1+(a/4*4);
}

class gcdex_t { int i; };

/***** precomputation of small primes ******/

int B1=10000000;
int B2=100001000;
int B3=1000000;

unsigned short *firstprimes;
int *primes;
int *rprimes;
int nbrprimes;

void initprimes() // precompute small primes
{
  printf("B1= %Ld  B2= %Ld\n",ll_t(B1),ll_t(B2));

  int tt=1+B2/8;
  unsigned char *tmp=(unsigned char*)malloc(tt);
  for(int i=0;i<tt;i++) tmp[i]=0xAA;
  tmp[0]=0xAE;
  for(long long i=3;i*i<=B2;i+=2)
    if((tmp[i/8]>>(i%8))&1) {
      for(long long j=i*i;j<=B2;j+=2*i) 
	tmp[j/8]&=~(1<<(j%8));
    }
  nbrprimes=0;
  for(long long i=2;i<=B2;i++)
    if((tmp[i/8]>>(i%8))&1)
      nbrprimes++;
  printf("%d primes <=B2\n",nbrprimes);
  primes=new int[nbrprimes];
  int j=1;primes[0]=2;
  for(int i=3;i<B2;i+=2) 
    if((tmp[i/8]>>(i%8))&1) {
      primes[j++]=i;
    }
  printf("init firstprimes OK\n");

  firstprimes=new unsigned short[B1];
  firstprimes[1]=0;
  for(int i=2;i<B1;i++) {
    if((tmp[i/8]>>(i%8))&1) {
      firstprimes[i]=0;
    } else {
      int j;
      for(j=0;primes[j]*primes[j]<=i;j++)
	if(i%primes[j]==0) {
	  firstprimes[i]=primes[j];
	  assert(primes[j]>1 && primes[j]<i);
	  break;
	}
      assert(primes[j]*primes[j]<=i);
    }
  }
  printf("init firstprimes2 OK\n");
  
  rprimes=new int[B3];
  for(int i=0;i<B3;i++) {
    rprimes[i]=-1;
  }
  int k=0;
  for(int i=2;i<B3;i++) {
    if((tmp[i/8]>>(i%8))&1) {
      rprimes[i]=k++;
    }
  }
  printf("init rprimes OK\n");
  
  free(tmp);
  assert(j==nbrprimes);
  printf("initprimes OK\n");
}

int nbrbaseisprime=5;

bool isprimei(int a)
{
  if(a<B1) return firstprimes[a]==0;
  return mpz_probab_prime_p(int_t(a).get_mpz_t(),nbrbaseisprime);
}

bool isprime(const int_t &a)
{
  if(a<B1) return firstprimes[toint(a)]==0;
  return mpz_probab_prime_p(a.get_mpz_t(),nbrbaseisprime);
}

#include "listfact.hpp"

/****************/

int BE=1000;
int_t LIMIT=1;
int LIMITNBRP=20;
int maxcrb=-1;


int_t assumedtrialcomp=
  int_t("100000000"); //option 'as'. (not checked with this program)

list<path_t> forceds; //forced paths ( 'fU' option )
list<path_t> forceds_fail;

int tabB[]={127,19,7,11,331,31,97,61,13,398581,1093,3,5,307,17,0};
int tabP[]={3,5,7,11,0}; //,13,17,0};
int tabC[]={0};
int *tab=tabB;

int tabstar[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}; //tab for "special" branching

int *tab_2=NULL;
int sizetab=0;
int sizetab_2=0;

int tab2(int i) // tab2(i) return the ith prime number in the our modified order
{
  if(i<sizetab_2) return tab_2[i];
  if(i+1>=nbrprimes) return 0;
  return primes[i+1];
}

bool isbefore(const int_t &p, int j) //returns true if p is before tab2(i) in our order 
{
  if(j>sizetab_2) return p<primes[j+1];
  if(p>=primes[sizetab_2+1]) return false;
  for(int i=0;i<j;i++)
    if(p==tab2(i)) return true;
  return false;
}

void inittab2()
{
  sizetab_2=0;
  assert(primes[0]==2);
  for(int i=0;tab[i];i++) {
    sizetab=i+1;
    while(primes[sizetab_2]<=tab[i]) sizetab_2++;
  }

  sizetab_2--;
  if(sizetab_2<=0) return;
  tab_2=new int[sizetab_2];
  int k=0;
  for(k=0;tab[k];k++) tab_2[k]=tab[k];
  for(int i=1;i<sizetab_2;i++) {
    int j;
    for(j=0;tab[j];j++) if(tab[j]==primes[i]) break;
    if(tab[j]) continue;
    tab_2[k++]=primes[i];
  }
  assert(k==sizetab_2);

  printf("sizetab_2=%d\n",sizetab_2);
}

#define B_BIG 1
#define B_SMALL 2
#define B_NOCMP 4

#define C_PROD 64
#define C_NBRP 128
#define C_COMP 256

#define F_DEF (C_PROD+B_BIG)
#define F_NBRP (C_NBRP+B_SMALL)
#define F_COMP (C_COMP+B_BIG)

#define F_COMPF (C_COMP+B_BIG+B_NOCMP)
#define F_FFF (C_PROD+B_BIG+B_NOCMP)
#define F_NBRPF (C_NBRP+B_SMALL+B_NOCMP)

int flags=F_DEF;

struct exp_t {
  exp_t(int m=0) {
    minexp=m;
    exp=0;
    prime=true;
    maxexp=-1;
  }
  
  bool prime;
  int minexp;
  int exp;
  int maxexp;
  
  int_t p;
  int q;
  
  bool special() const {return minexp%2;}
  bool br() const {return minexp!=0;}
};

struct pfact_t{
  pfact_t() {}

  pfact_t(int_t a) {
    fs[a]=exp_t(0);
  }
  
  map<int_t,exp_t> fs;
  void aff(ostream &o) const {
    for(map<int_t,exp_t>::const_iterator it=fs.begin();it!=fs.end();it++) {
      if(!it->second.prime) o<<"c";
      o<<it->first<<"^"<<it->second.exp<<" ";
    }
    o<<endl;
  }

  int_t prod() const {
    int_t r=1;
    for(map<int_t,exp_t>::const_iterator it=fs.begin();it!=fs.end();it++)
      r*=pow(it->first,it->second.exp);
    return r;
  }

  void aff2(ostream &o) const {
    for(map<int_t,exp_t>::const_iterator it=fs.begin();it!=fs.end();it++) {
      if(!it->second.prime) o<<"c";
      o<<it->first<<"^("<<it->second.minexp;
      if(it->second.maxexp==it->second.minexp) o << "*";
      o<<","<<it->second.exp<<") ";
      if(!it->second.prime) o<<"["<<it->second.p<<","<<it->second.q<<"] ";
    }
    o<<endl;
  }

  bool addprime(const int_t &a,int e=1) {
    if(fs[a].maxexp>=0 && (fs[a].exp+e)>fs[a].maxexp) return false;
    fs[a].exp+=e;
    return true;
  }

  void addcomp(const int_t &a,const int_t &p,int q) {
    if(fs[a].exp) {
      cerr << " FATAL double comp "<< a << endl;
      aff2(cerr);
      abort();
    }
    fs[a].exp+=1;
    fs[a].prime=false;
    fs[a].p=p;
    fs[a].q=q;
  }

  // return true if we have already supposed a special factor
  bool special() const {
    map<int_t,exp_t>::const_iterator it;
    for(it=fs.begin();it!=fs.end();it++) 
      if(it->second.special()) return true;
    return false;
  }

  // check if gcd of every pair of factor is 1. just an assertion
  void assertgcd() const {
    map<int_t,exp_t>::const_iterator it,it2;
    for(it=fs.begin();it!=fs.end();it++)
      for(it2=fs.begin();it2!=it;++it2) {
	int_t g;
	mpz_gcd(g.get_mpz_t(),it->first.get_mpz_t(),it2->first.get_mpz_t());
	if(g!=1) {
	  cerr << "FATAL gcd = " << g << " : " << it->first << " " << it2->first << endl;
	  aff2(cerr);
	  {
	    ofstream gcdout("gcdfacts.txt",ios_base::out|ios_base::app);
#ifdef NOTH
	    if(envprof)
	      gcdout << "# " << envprof->indent() << endl;
#endif
	    if(!it->second.prime)
	      gcdout <<it->second.p<<" "<<it->second.q+1 << " " << g << endl;
	    if(!it2->second.prime)
	      gcdout <<it2->second.p<<" "<<it2->second.q+1 << " " << g << endl;
	  }
	  throw gcdex_t();
	  //abort();
	}
      }
  }


  // add a list of factors with exponents. return false if this is impossible (in "special" branchings)
  bool add(const pfact_t &o) {
    map<int_t,exp_t>::const_iterator it;
    for(it=o.fs.begin();it!=o.fs.end();it++) {
      if(fs[it->first].maxexp>=0) {
	if(fs[it->first].exp+it->second.exp>fs[it->first].maxexp) return false;
      }
      fs[it->first].exp+=it->second.exp;
      if(!it->second.prime) {
	fs[it->first].prime=false;
	fs[it->first].p=it->second.p;
	fs[it->first].q=it->second.q;
      }
    }

    assertgcd();
    return true;
  }


  /******* CHECKS **********/

  int nbrp() const {
    bool sp=special();
    int r=0,exp;
    int_t mi=0;
    
    map<int_t,exp_t>::const_iterator it;
    if(sp==false) {
      for(it=fs.begin();it!=fs.end();it++) {
#ifdef DOUBLECHECK
	assert(it->second.minexp%2==0);
#endif
	if(it->second.prime==false || (it->first%4==1 && it->second.minexp==0)) { ////celui là peut etre impair
	  r--;break; 
	}
	//opt?: si composite, r-- et break aussi
      }
      //if(it==fs.end()) r++; // il faut un autre premier special 
    }

    for(it=fs.begin();it!=fs.end();it++) {
      if(it->first==2) continue;
      if(it->second.prime) {
	if((it->second.minexp%2)==0) {
	  exp=nexteven(it->second.exp);
	  if(exp<2) exp=2;
	  if(exp<it->second.minexp) exp=it->second.minexp;
	} else {
#ifdef DOUBLECHECK
	  assert((it->second.minexp%2)==1);
	  //assert(it->second.exp<2);
#endif
	  exp=next1mod4(it->second.exp);
	  if(exp<it->second.minexp) exp=it->second.minexp;
	}
      } else {
	if(isperfectpower(it->first)) {
	  //if it is a perfect power, we cannot assume exp>=4 (but this is very improbable).
	  cerr << "FATAL: " << it->first << "is a perfect power\n";
	  abort();
	}

	exp=4; 
      }
      
      r+=exp;
    }
    return r;
  }

  bool checkNBR() const {
    return nbrp()<LIMITNBRP;
  }

  /************ COMP ****************/

  bool checkCOMP() const {   // marche pas avec 'all'
    bool sp=special();

    map<int_t,exp_t>::const_iterator it;
    
    for(it=fs.begin();it!=fs.end();it++) { 
      if(it->second.prime && it->first>2) {
	int exp=0;
	if(it->second.exp>it->second.minexp) {
	  exp=it->second.exp;
	  if(sp || it->first%4!=1) exp=nexteven(exp);
	  else {
	    if(nexteven(exp)<next1mod4(exp)) exp=nexteven(exp);
	    else exp=next1mod4(exp);
	  }
	} else exp=it->second.minexp;
	//printf("e=%d m=%d (sp=%d) (mod=%d)-> %d\n",it->second.exp,it->second.minexp,sp,(it->first%4)==1?1:3,exp);
	assert(exp>0);
	if(pow(it->first,exp)>LIMIT) return false;
      }
    }
    return true;
  }

  /************** B ***************/

  int_t B() const { // returns a lower bound for N
    int_t r=1;
    int_t spp=1;
    int_t mdfsp=1,mdfnsp=1;


    bool sp=special();
    map<int_t,exp_t>::const_iterator it;

    for(it=fs.begin();it!=fs.end();it++) {
      if(it->first==2) continue;
      int exp=0;
      int expsp=0;
      int expnsp=0;
      if(it->second.prime) {
	if(it->second.minexp==0) {
	  if(it->first%4==1) {
	    expnsp=nexteven(it->second.exp);
	    expsp=next1mod4(it->second.exp);
	  } else {
	  exp=nexteven(it->second.exp);
	  }
	} else if(it->second.minexp%2==0) {
	  exp=nexteven(it->second.exp);
	  if(exp<it->second.minexp) exp=it->second.minexp;
	} else { // (it->second.minexp%2)==1
	  assert(sp);
#ifdef DOUBLECHECK
	  assert(it->second.minexp%4==1);
	  assert(it->first%4==1);
#endif
	  exp=next1mod4(it->second.exp);
	  if(exp<it->second.minexp) exp=it->second.minexp;
	}

      } else {
	expsp=expnsp=1; // le composite peut avoir des carrés...
	if(!(it->second.exp==1 && it->second.minexp==0)) {
	  cerr << "FATAL comp " << it->first << " minexp=" <<it->second.minexp<< " exp=" << it->second.exp<<endl;
	}
#ifdef DOUBLECHECK
	assert(it->second.exp==1 && it->second.minexp==0);
#endif
	//if(exp<it->second.minexp) exp=it->second.minexp;
      }

      if(exp) {
	  r*=pow(it->first,exp);
      } else {
	int_t tcsp=pow(it->first,expsp);
	int_t tcnsp=pow(it->first,expnsp);

	
	if(tcnsp>spp) {
	  spp=tcnsp;
	}

	if(tcsp*mdfnsp<tcnsp*mdfsp) {
	  mdfsp=tcsp;
	  mdfnsp=tcnsp;
	}

	r*=tcnsp;
      }
    }

    
      if(sp) {
	return r;
      }
      assert(mdfnsp>=mdfsp);
      return r*mdfsp/mdfnsp;


  }

  bool checkB() const {
    return (B()<LIMIT);
  }

  /****** AB ******/
  
  mpq_class  compAb() const { 

    int_t ma=0;
    bool sp=special();
    map<int_t,exp_t>::const_iterator it;

    mpq_class ab=1;
    
    for(it=fs.begin();it!=fs.end();it++) {
      if(it->first<=2 || !it->second.prime) continue;
      
      int exp=0;
      if(it->second.minexp) {
	if((it->second.minexp%2)==0)
	  exp=nexteven(it->second.exp);
	else
	  exp=next1mod4(it->second.exp);
      } else {
	exp=nexteven(it->second.exp);
      }
      if(exp<it->second.minexp) exp=it->second.minexp;
      
      ab*=sigma(it->first,exp);
      ab/=pow(it->first,exp);
      
      if(sp && ab>2)
	return ab;
    }
    if(sp) return ab;

    mpq_class abw=ab;
    for(it=fs.begin();it!=fs.end();it++) { 
      if(it->first<=2 || false==it->second.prime || it->second.minexp!=0 || 1!=(it->first%4)) continue;
      
      int oexp=nexteven(it->second.exp);
      int nexp=next1mod4(it->second.exp);
      mpq_class ab2=ab;
      
      ab2/=sigma(it->first,oexp);
      ab2*=pow(it->first,oexp);
      
      ab2*=sigma(it->first,nexp);
      ab2/=pow(it->first,nexp);
      
      if(ab2<abw) abw=ab2;
    }

    return abw;
  }

  mpq_class Ab(bool worst=false) const {
    mpq_class r;
    mpq_class rc;
    if(!worst)
      return compAb();

    map<int_t,exp_t>::const_iterator it;

    r=1;rc=1;
    for(it=fs.begin();it!=fs.end();it++) {
      if(it->first>2) {
	if(it->second.prime) {
	  mpq_class a;
	  if(it->second.maxexp==-1)
	    a=1+mpq_class(1)/(it->first-1);
	  else
	    a=mpq_class(sigma(it->first,it->second.maxexp))/pow(it->first,it->second.maxexp);
	  r*=a;
	} else {
	  int_t l=1;
	  int k=0;
	  while(l<=it->first) {k++;l*=assumedtrialcomp;}
	  mpq_class a=1+mpq_class(1)/assumedtrialcomp;
	  for(;k>0;k--) rc*=a;
	}
      }
    }
    
    return r*rc;
  }

  bool checkAb() const {
    return compAb()<=2;
  }


  /****** BESTS *********/

  void branchables(set<int_t> &s) const {
    s.clear();
    map<int_t,exp_t>::const_iterator it;
    for(it=fs.begin();it!=fs.end();it++) {
      if(it->first!=2 && (!it->second.br()) && it->second.prime) 
	s.insert(it->first);
    }
  }

  void bests(list<int_t> &s) const {
    set<int_t> s2;
    branchables(s2);
    s.clear();
    for(set<int_t>::iterator it=s2.begin();it!=s2.end();++it)
      s.push_front(*it);
  }

  void bestssmalls(list<int_t> &s) const {
    set<int_t> s2;
    branchables(s2);
    s.clear();
    for(set<int_t>::iterator it=s2.begin();it!=s2.end();++it)
      s.push_back(*it);
  }
};

string outrbcompname="";
string prefixname="";



void initoracle(int &ac, char **&av)
{
  initprimes();
  initlistfactors();

  for(int i=1;i<ac;i++) {
    fprintf(stderr,"read %s\n",av[i]);
    readfactors(av[i]);
  }
}



int limsize=0;
int profmax=0;

int upto=0;

/// thread encapsulation
struct th_t {

  const path_t *envprof;

  ostream &indcout() {
    if(!envprof) return tout;
    return (tout << envprof->indent());
  }

  int id; //thread id;

  // "tout" is the ostream where we write the tree.
#ifndef NOTH
  ofstream tout;
#else 
  ostream &tout;
#endif

  static string id2str(int id) {
#ifndef NOTH
    char b[10];
    snprintf(b,10,"%s%d_",prefixname.c_str(),id);
    return b;
#else
    return prefixname;
#endif
  }
  
  th_t(int i=0):id(i),
#ifndef NOTH
		tout((id2str(i)+"log.txt").c_str(),ios_base::out|ios_base::app) 
#else
		tout(cout)
#endif
  {
    stat_rec=0;
    envprof=NULL;
    outrbcomp=NULL;
    if(outrbcompname!="") {
      outrbcomp=new ofstream((id2str(i)+outrbcompname).c_str(),ios_base::out|ios_base::app);
      *outrbcomp<<"#"<<cmdline<<endl;
    }
  }

  ~th_t() {
    tout<< "thread end. stat_rec=" << stat_rec <<endl;
    if(outrbcomp) delete outrbcomp;
  }

/* check for contracitions: abundancy and limit */
bool checkcontradiction(const pfact_t &fsp)
{
  if((flags & C_PROD) &&!fsp.checkB()) {
    if(verb>=2) indcout()<< "N>LIMIT :\n";
    return true;
  }
  
  if((flags & C_NBRP) &&!fsp.checkNBR()) {
    if(verb>=2) indcout()<< "NBRP>LIMIT :\n";
    return true;
  }
  
  if((flags & C_COMP) &&!fsp.checkCOMP()) {
    if(verb>=2) indcout()<< "COMP>LIMIT\n";
    return true;
  }
  
  if(!fsp.checkAb()) {
    if(verb>=2) indcout()<< "Ab>2 :\n";
    return true;
  }
  
  return false;
}

/* check for contracitions: previous + special banching and forb factors */
bool checkcontradiction2(pfact_t &fsp, const pfact_t &fs, const pfact_t &fa,ll_t borne)
{
  fsp=fs;
  
  if(!fsp.add(fa)) {
    if(verb>=2) indcout()<< "contradiction add :\n";
    return true;
  }

  if(checkcontradiction(fsp)) return true;
  
  bool skip=false;
  map<int_t,exp_t>::const_iterator it;
  for(it=fa.fs.begin();it!=fa.fs.end();it++) {
    if(it->second.prime && it->first>2 && isbefore(it->first,borne)) {
      if(fs.fs.find(it->first)!=fs.fs.end()) continue; // not a new prime factor
      if(verb>=2) indcout()<< it->first << " already eliminated:\n";
      skip=true;
      break;
    }      
  }

  if(skip) return true;

  return false;
}


path_t forced; //forced path ( 'U' option )
ofstream *outrbcomp;

ull_t stat_rec;

int th_id;


// 'find' factors of f (factor if s(p^q)), and put the known decomposition into pfact
int oraclefact(const int_t &p,int q,pfact_t &pfact,int_t &f) 
{
  path_t prof;
  if(envprof) prof=*envprof;
  
  facts_t *fa=getfacts(p,q);
  
  bool ok=0;

  facts_t::const_iterator it;
  for(it=fa->begin();it!=fa->end();++it)
    while(isdiv(f,*it)) {
      ok=1;
      pfact.addprime(*it);
      f/=*it;
    }

  if(f==1) return 1;
  
  if(isprime(f)) {
    pfact.addprime(f);
    return 1;
  }
  
  pfact.addcomp(f,p,q);
  
  if(!ok) return 0;
  return 2;
}

#if 0
bool rec(pfact_t &fs,int_t p,const path_t &prof,ll_t borne);
#endif

bool rec_range(pfact_t &fs, int cond,const path_t &prof,ll_t borne)
{
  const path_t *oprof=envprof;
  envprof=&prof;

  if(prof.h()<limsize) {
    prof.aff(tout); 
    if(prof.h()==limsize-1) tout << " BORDERPATH";
    tout <<endl;
  }
  ll_t ostat=stat_rec;

  if(maxcrb>=0 && prof.crb()>=maxcrb) {
    if(verb>=1) 
      indcout() << "REC_RANGE FAIL MAXCRB\n";
    envprof=oprof;
    return false;
  }

  mpq_class abw=fs.Ab(true);

  if(abw>=2) {
    if(verb>=1) 
      indcout() << "REC_RANGE FAIL abw>=2\n";
    envprof=oprof;
    return false;
  }

  int rlnbr=0;
  if(cond&C_NBRP) { 
    rlnbr=LIMITNBRP-fs.nbrp();
    rlnbr=(rlnbr+1)/2;
  }

  int_t dB=0;
  if(cond&C_PROD) {
    dB=LIMIT/fs.B();
    if(verb>=2) 
      indcout() << "dB= "<< dB <<endl;
  }

  ll_t pmax;

  if(cond) {
    int i;
    for(i=0;i<nbrprimes;i++) {
      ull_t p=primes[i];
      int_t pt=ll2mpz(p);
      if(cond&C_NBRP) {
	mpq_class a=1+mpq_class(1)/pt;
	mpq_class b=abw;
	for(int i=0;i<rlnbr;i++)
	  b*=a;
	if(b<2) break;
      } else 
      if(cond&C_PROD) {
	int q;
	int_t r=1;
	for(q=1;;q++) {
	  r*=pt;
	  if(r>dB) break;
	}
	q=q/2;
	mpq_class b=abw;
	mpq_class a=1+mpq_class(1)/(pt-1);
	for(int i=0;i<q;i++) {
	  b*=a;
	  if(b>=2) break;
	}
	if(b<2) {
	  if(verb>=1) 
	    indcout() << "RANGE stop p=" << pt << " q=" <<q << " ab= "<< b.get_d() << endl;
	  break;
	}
      } else {
	abort();
      }
    }
    pmax=primes[i]-1;
  }

  ull_t str=stat_rec;

  if(verb>=1) 
    indcout() << "REC_RANGE N="<< prof.crb()<< " npmax="<< pmax<< " borne="<< borne << endl;

  bool rok=true;

  ull_t p;
  for(int i=0;;i++) {
    
    p=tab2(i);
    if(p==0) break;
    assert(p>2);

    int_t pt=ll2mpz(p);

    if(p>pmax && i>=sizetab) {
      if(verb>=4) 
	indcout() << p << " > pmax and not in S\n";
      if(i>sizetab_2) break;
      continue;
    }

    if(i<borne) {
      if(verb>=3) 
	indcout()<< p << " already eliminated\n";
      continue;
    }
    
    map<int_t,exp_t>::const_iterator it=fs.fs.find(pt);

    if(it!=fs.fs.end() && it->second.minexp) {
      if(verb>=2) 
	indcout() << "already "<< p<<endl;
      continue;
    }

    {
      //tout << "RANGE: " << p <<endl;
      assert(i>=borne);
      pfact_t fs2=fs;
      path_t prof2=prof;
      prof2.add(-1);
      bool r=rec(fs2,pt,prof2,i);
      if(!r) rok=false;
    }
  }

  if(p==0) {
    if(verb>=1) 
      indcout() << "REC_RANGE FAIL p limit"  << endl;
    rok=false;
  }

  if(verb>=1) 
    indcout() << "\\REC_RANGE " << rok << " N="<< prof.crb() << " size=" << stat_rec-str<<endl;

  envprof=oprof;

  if(prof.h()<limsize) {
    prof.aff(tout); 
    tout << " [ size= " << stat_rec-ostat<< " ret= " << rok <<" ]\n";
  }

  return rok;
}


bool rec(pfact_t &fs,int_t p,const path_t &prof,ll_t borne)
{
  if(profmax && prof.h()>=profmax) return false;

  int forcedq=-1;
  if(forced.h()>prof.h()) {
    if(!forced.compat(prof,p)) {
      //indcout()<< "wrong forced path !!!\n";
      return true;
    }

    list<pair<int_t,int> >::const_iterator it1=forced.p.begin(),it2=prof.p.begin();
    while(true) {
      if(it1==forced.p.end() || it2==prof.p.end()) break;
      it1++;it2++;
    }
    assert(it1!=forced.p.end());
    forcedq=it1->second;
  }

  if(prof.h()<limsize) {
    prof.aff(tout); 
    if(prof.h()==limsize-1) tout << " BORDERPATH";
    tout <<endl;
  }
  ll_t ostat=stat_rec;

  stat_rec++;

  const path_t *oprof=envprof;
  envprof=&prof;

  int q=0;
  bool ok=true;
  try {
  while(1) {
    fs.fs[p].minexp=0;
    q++;
    if((q&1) && (fs.special() || (p%4)!=1))
      continue;

    int star=0;
    if(p<sizeof(tabstar)/sizeof(tabstar[0])) star=tabstar[toint(p)];
    //if(q%2==0 && (p!=3 || q!=8) && firstprimes[q+1]) continue;
    assert(!(star!=0 && star!=2 && star!=4 ));
    assert(!(star==4 && (p%4)==1));
    if(star==0) {
      if(firstprimes[q+1]) continue;
    }
    if((star==2&&(p%4)!=1) && q!=8 && firstprimes[q+1]) continue;
    if((star==2&&(p%4)==1) && q!=8 && q!=5 && firstprimes[q+1]) continue;
    if((star==4) && q!=8 && q!=14 && q!=24 && firstprimes[q+1]) continue;
    fs.fs[p].maxexp=-1;
    
    fs.fs[p].minexp=q;

    if(q<=star) fs.fs[p].maxexp=q;
    if(q==1) fs.fs[p].maxexp=q;

    if(forcedq>=0) {
      if(q<forcedq) continue;
      if(q>forcedq) break;
    }

#if 0
    if(!forced.compat(prof,p,q)) { //for 'U' option

      /* contradictions */
      if(q>=2 && checkcontradiction(fs)) {
	tout<<"break without process full path...\n";

	break;
      }
      
      continue; 
    }
#endif

    if(verb>=2) indcout()<<"~ "<< p << " " << q <<endl;

    /* "special" branching contradiction */
    if(fs.fs[p].maxexp>=0 && fs.fs[p].exp>fs.fs[p].maxexp) {
      if(verb>=2)
	indcout()<<" exp>max"<<endl;
      continue;
    }

    /* contradictions */
    if(checkcontradiction(fs)) {
      if(q>=2) break;
      continue;
    }

    int_t bs=sigma(p,q);
    int_t b;

    if(verb>=3) {
      indcout()<<"current: ";fs.aff2(tout);
      indcout()<<"s("<<p<<"^"<<q<<")= "<<bs<<endl;
    }


    pfact_t fa,fsp;
    {
      int rf;
      bool skip=false;


      b=bs;
      rf=oraclefact(p,q,fa,b);

#ifdef DOUBLECHECK
      if(fa.prod()!=bs) {
	cerr << "assert prod error: "<<bs << " != "; fa.aff(cerr); tout<<endl;
	abort();
      }
#endif
      
      if(checkcontradiction2(fsp,fs,fa,borne)) {
	skip=1;
      }
      

      if(verb>=2) {
	indcout()<<"s("<<p<<"^"<<q<<")= ";fa.aff(tout);
	indcout()<<"c: ";fsp.aff2(tout);
      }
      
      if(skip) continue;
      
      if(rf!=1) {
	if(verb>=3)
	  indcout()<<"cannot fact "<<p << " " <<q+1<<" "<< b <<endl;
	
	if(flags&B_NOCMP) { //Pascal's favourite option
	  ok=false;
	  continue;
	}
      }
    }

    list<int_t> lb;
    if(flags&B_BIG) fsp.bests(lb);
    else if(flags&B_SMALL) fsp.bestssmalls(lb);
    else abort();


    if(outrbcomp && !lb.size()) {
      for(map<int_t,exp_t>::const_iterator it=fsp.fs.begin();it!=fsp.fs.end();it++) {
	if(it->second.prime) continue;
	(*outrbcomp)<<it->second.p<<" "<<it->second.q<<" "<<it->first<<endl;
      }
    }

    /* if no factor, try CRB */
    
    if(!lb.size()) {


      bool r=false;
      if(verb>=1) {
	indcout()<< "no branch ! : "; // << endl;
	fsp.aff2(tout);
      }	

      { //aff
	mpq_class ab=fsp.Ab(),abw=fsp.Ab(true);
	double fl=ab.get_d(),flw=abw.get_d();
	if(verb>=1) 
	  indcout() << "Ab= " << fl << " Abw=" << flw << endl;
      }

      if(flags & C_NBRP) {
	{ //aff 
	  mpq_class ab=fsp.Ab(),abw=fsp.Ab(true);
	  double fl=ab.get_d(),flw=abw.get_d();
	  int re=LIMITNBRP-fsp.nbrp();
	  int np=(re+1)/2;
	  assert(np>0);
	  double lala=2/fl;
	  double lim=pow(lala,1./np);
	  assert(lim>1.);
	  double lp=1./(lim-1.);
	  if(verb>=1) 
	    indcout() << "P max: " << lp << endl; 
	  if(flw<2) {
	    double lala=2/flw;
	    double lim=pow(lala,1./np);
	    assert(lim>1.);
	    double lp=1./(lim-1.);
	    if(verb>=1) 
	      indcout() << "P max w: " << lp << endl; 
	  }
	}

	// CRB:
	path_t p2=prof;
	p2.add(p,q);
	r=rec_range(fsp,C_NBRP,p2,borne);

	if(r==false) ok=false;
	if(forced.h()>= p2.h()) break;
	
      } else 
      if(flags & C_PROD) {
	// CRB:
	path_t p2=prof;
	p2.add(p,q);
	r=rec_range(fsp,C_PROD,p2,borne);

	if(r==false) ok=false;
	if(forced.h()>= p2.h()) break;

      } else abort();


      continue;
    }

    /* if factor(s), branch on the "best" prime */

    list<int_t>::const_iterator it=lb.begin();
    path_t p2=prof;
    p2.add(p,q);
    if(!rec(fsp,*it,p2,borne)) ok=false;

    if(forced.h()>= p2.h()) break;
  }
  } catch(gcdex_t e) {
    indcout()<< "GCD error..."<<endl;
    ok=false;
  }

  if(verb>=2)
    indcout()<<"end " << p << " " << q << endl;

  envprof=oprof;

  if(prof.h()<limsize) {
    prof.aff(tout); 
    tout << " [ size= " << stat_rec-ostat<< " ret= " << ok <<" ]\n";
  }
  
  return ok;
}

bool main_proc(path_t forced_,int uniq=0, int after=0)
{
  bool ret=1;
  forced=forced_;
  if(forced.h()) {
    tout << "*** "; forced.aff(tout); tout << " ***\n";
    tout.flush();
    uniq=toint(forced.p.begin()->first);
    after=0;
  }
  for(int mit=0;upto || tab[mit] ;mit++) {
    int i=tab2(mit);
    assert(i>2);
    if(upto && i>upto) break;
    if(uniq && i!=uniq) continue;
    if(after) uniq=0;
#ifdef NOTH
    if(mit<100 || mit%5000==0) 
      fprintf(stderr,"i=%d        \r",tab2(mit));
#endif
    path_t p;
    pfact_t pf = pfact_t(int_t(i));
    ll_t osr=stat_rec;
    bool rr=rec(pf,i,p,mit);
    if(!rr) ret=0;
    tout << "\n--- RETURN FOR " << i;
    if(forced.h()) {tout<<" / "; forced.aff(tout);}
    tout<< " : " << rr<< " (ar : " << ret << " size= " <<stat_rec-osr<<" ) "<<endl;
    tout.flush();
  }
  
  tout<< "stat_rec=" << stat_rec <<endl;
  return ret;
}

bool main_proc_uniq(int ii)
{
  forced=path_t();

  tout << "*** "<< tab2(ii) <<  " ***\n";
  tout.flush();
  
  path_t p;
  pfact_t pf = pfact_t(int_t(tab2(ii)));
  bool rr=rec(pf,tab2(ii),p,ii);
  tout << "\n--- RETURN FOR " << tab2(ii) << " : " << rr<<endl;
  tout.flush();
  return rr;
}

};

#ifndef NOTH

pthread_t thtab[128];
pthread_mutex_t mutex;
int thmode=0;
int thmit=0;
int thupto=0;

void *thread_main(void *pt)
{
  int id=*(int*)pt;
  th_t th(id);
  
  printf("START THREAD %d\n",id);fflush(stdout);

  bool ret=1;
  while(1) {
    path_t f;
    bool end=false;

    pthread_mutex_lock(&mutex);

    int ip=-1;
    if(thmode==0) {
      if(forceds.empty()) end=true;
      else {f=forceds.front();forceds.pop_front();}
    } else {
      if(thupto || tab[thmit]!=0) {
	ip=thmit;
	int p=tab2(thmit);
	assert(p>2);
	if(thupto && p>thupto) ip=-1;
      }
      if(ip>=0) thmit++;
      else end=true;
    }
    
    pthread_mutex_unlock(&mutex);
    
    if(end) break;
    
    int rr;
    if(thmode==0)
      rr=th.main_proc(f);
    else 
      rr=th.main_proc_uniq(ip);
    if(!rr) {
      ret=0;
      pthread_mutex_lock(&mutex);
      forceds_fail.push_back(f);
      pthread_mutex_unlock(&mutex);
    }
  }

  printf("END THREAD %d RET=%d\n",id,ret);fflush(stdout);

  return NULL;
}
#endif

int main(int ac, char **av)
{
  path_t forced;
  int uniq=0,after=0;
  int nbrthreads=1;


  {
    strcpy(cmdline,"");
    for(int i=0;i<ac;i++) {
      if(i) strcat(cmdline," ");
      strcat(cmdline,av[i]);
    }
    printf("cmd line: %s\n",cmdline);
  }


  int addflags=0;
  while(ac>1 && av[1][0]=='-') {
    if(0==strcmp(av[1]+1,"v")) {verb++;ac-=1;av+=1;continue;}
    if(0==strcmp(av[1]+1,"q")) {verb--;ac-=1;av+=1;continue;}

    if(0==strcmp(av[1]+1,"SIZE")) {BE=atoi(av[2]);ac-=2;av+=2;continue;}
    if(0==strcmp(av[1]+1,"NBRP")) {
      tab=tabP;      
      flags=F_NBRP;
      LIMITNBRP=atoi(av[2]); 
      assert(LIMITNBRP%2);
      ac-=2;av+=2;
      continue;
    }

    if(0==strcmp(av[1]+1,"COMP")) {
      tab=tabC;
      flags=F_COMP;
      BE=atoi(av[2]);
      ac-=2;av+=2;continue;
    }

    if(0==strcmp(av[1]+1,"u")) {uniq=atoi(av[2]); ac-=2;av+=2;continue;}
    if(0==strcmp(av[1]+1,"a")) {uniq=atoi(av[2]); after=1;ac-=2;av+=2;continue;}
    if(0==strcmp(av[1]+1,"U")) {
      int_t p=int_t(av[2]);
      int q=atoi(av[3]);
      forced.add(p,q);
      av+=3;ac-=3;
      continue;
    }

    if(0==strcmp(av[1]+1,"pm")) {profmax=atoi(av[2]);ac-=2;av+=2;continue;}
    if(0==strcmp(av[1]+1,"lm")) {limsize=atoi(av[2]);ac-=2;av+=2;continue;}

    if(0==strcmp(av[1]+1,"t")) {nbrthreads=atoi(av[2]);ac-=2;av+=2;continue;}

    if(0==strcmp(av[1]+1,"as")) {assumedtrialcomp=int_t(av[2]);ac-=2;av+=2;continue;}

    if(0==strcmp(av[1]+1,"fileprefix")) {prefixname=(av[2]);ac-=2;av+=2;continue;}

    if(0==strcmp(av[1]+1,"fU")) {
      FILE *in=fopen(av[2],"rt");
      assert(in);
      char bf[100000];
      while(fgets(bf,100000,in)) {
	assert(strlen(bf)<99900);
	if(bf[0]=='#' || bf[0]==0 || bf[0]=='\n') continue;
	forceds.push_back(path_t(bf));
      }
      fclose(in);
      av+=2;ac-=2;
      continue;
    }

    if(0==strcmp(av[1]+1,"F")) {addflags|=B_NOCMP;ac-=1;av+=1;continue;}


    if(strcmp(av[1]+1,"crb")==0) {maxcrb=atoi(av[2]);ac-=2;av+=2;continue;}

    if(0==strcmp(av[1]+1,"ut")) {
      double lala=atoi(av[2]);
      upto=pow(10.,lala);
      printf("upto %d\n",upto);
      ac-=2;av+=2;continue;
    }

    if(0==strcmp(av[1]+1,"UT")) {
      upto=atoll(av[2]);
      printf("upto %d\n",upto);
      ac-=2;av+=2;continue;
    }
    
    if(0==strncmp(av[1]+1,"X",1)) {
      int k=atoi(av[1]+2);
      int i=atoi(av[2]); 
      assert(k==2 || k==4);
      assert(i<sizeof(tabstar)/sizeof(tabstar[0]));
      tabstar[i]=k;
      ac-=2;av+=2;
      continue;
    }

    if(strcmp(av[1]+1,"rbcomp")==0) {
      printf("rbcomp to %s\n",av[2]);
      outrbcompname=av[2];
      //outrbcomp=new ofstream(av[2],ios_base::out|ios_base::app);
      ac-=2;av+=2;
      continue;
    }

       
    printf("bad option: %s\n",av[1]);
    exit(1);
  }
  flags|=addflags;

  printf("list of factors: ");
  for(int i=0;tab[i];i++) {printf("%d ",tab[i]);}
  printf("\n");
  printf("\n");

  LIMIT=1;
  for(int i=0;i<BE;i++)
    LIMIT*=10;

  nbrbaseisprime =2; // only 2 bases for a quick check when read facts 

  initoracle(ac,av);

  makesmallfacts(3,8); makesmallfacts(3,14); makesmallfacts(3,24);
  makesmallfacts(5,5); makesmallfacts(5,8);
  makesmallfacts(7,8);
  makesmallfacts(11,8);
  makesmallfacts(13,5); makesmallfacts(13,8);
  makesmallfacts(17,5); makesmallfacts(17,8);
  makesmallfacts(19,8);
  makesmallfacts(23,8);    

  inittab2();

  nbrbaseisprime=5;

  if(forced.h())
    forceds.push_front(forced);

    
    printf("forceds.size= %d\n",int(forceds.size()));

#ifdef NOTH

  if(forceds.empty()) forceds.push_back(path_t());

  printf("START\n");
  fprintf(stderr,"START      \r");

  {
    th_t th;
    
    for(list<path_t>::iterator it=forceds.begin();it!=forceds.end();++it) {
      th.main_proc(*it,uniq,after);
    }
  }
#else

  if(forceds.empty()) {
    thmode=1;
    assert(uniq==0 || after!=0);
    if(uniq) {
      for(thmit=0;tab2(thmit) && tab2(thmit)!=uniq;thmit++);
      assert(tab2(thmit));
    }
    if(upto) thupto=upto;
  }

  printf("START\n");fprintf(stderr,"START      \r");

  if(nbrthreads<=1) {
    int zero=0;
    thread_main(&zero);
  } else {
    int tid[nbrthreads];
    for(int i=0;i<nbrthreads;i++) {
      tid[i]=i;
      pthread_create(&thtab[i],NULL,thread_main,tid+i);
    }
    
    printf("WAIT...\n");fflush(stdout);

    for(int i=0;i<nbrthreads;i++) {
      pthread_join(thtab[i],NULL);
      printf("JOIN THREAD %d\n",i);fflush(stdout);
    }
  }

#endif


  printf("END\n");fflush(stdout);
  fprintf(stderr,"END\n");fflush(stderr);


  return 0;
}
